# -*- coding: utf-8 -*-

"""Integration and functional tests for ``pyramid_twitterauth``."""

import unittest

try: # pragma: no cover
    from webtest import TestApp, TestRequest, TestResponse
except ImportError: # pragma: no cover
    pass

def config_factory(is_authenticated=False, **settings):
    """Call with settings to make and configure a configurator instance, binding
      to an in memory db.
    """
    
    from pyramid.config import Configurator
    from pyramid.session import UnencryptedCookieSessionFactoryConfig
    
    # Patch the settings to use an in memory db for testing, which should
    # be dropped every time the app is created.
    settings['sqlalchemy.url'] = 'sqlite:///:memory:'
    settings['basemodel.should_drop_all'] = True
    # Patch the settings to use a dummy app to connect to Twitter.
    key = 'xZNXRgs8oqEcHOBxf4vpw'
    secret = 'E9G2yBiyvp16rZ9hh3QhOe5MXU1Ecwo2LIYEbLbHdto'
    settings['twitterauth.oauth_consumer_key'] = key
    settings['twitterauth.oauth_consumer_secret'] = secret
    if is_authenticated:
        settings['simpleauth.set_default_permission'] = False
    # Initialise the ``Configurator`` and setup a session factory.
    config = Configurator(settings=settings)
    config.set_session_factory(UnencryptedCookieSessionFactoryConfig('psst'))
    # Include the dependencies.
    config.include('pyramid_tm')
    config.include('pyramid_simpleauth')
    config.commit()
    config.include('pyramid_twitterauth')
    config.include('pyramid_basemodel')
    # Fake authentication if need be.
    if is_authenticated:
        config.registry.settings['simpleauth.set_default_permission']
        from pyramid_simpleauth.model import User
        return_true = lambda request: True
        return_user = lambda request: User()
        config.set_request_property(return_true, 'is_authenticated', reify=True)
        config.set_request_property(return_user, 'user', reify=True)
    # Return the configurator instance.
    return config

class TestUnpackCallback(unittest.TestCase):
    """Integration tests for ``view._unpack_callback``."""
    
    def setUp(self):
        from mock import Mock
        mock_request = Mock()
        mock_request.GET = {}
        mock_request.session = {
            'twitter_request_token_key': 'key',
            'twitter_request_token_secret': 'secret',
        }
        self.request = mock_request
        self.request.route_url.return_value = '/'
        # Setup mock oauth handler and tweepy client.
        self.mock_handler_factory = Mock()
        self.mock_handler = Mock()
        self.mock_handler.access_token.key = 'key'
        self.mock_handler.access_token.secret = 'secret'
        self.mock_handler_factory.return_value = self.mock_handler
        self.mock_api_factory = Mock()
        self.mock_client = Mock()
        self.mock_client.last_response.getheader.return_value = 'read'
        self.mock_api_factory.return_value = self.mock_client
    
    def makeOne(self, *args, **kwargs):
        """Call the OAuth callback view and return the response."""
        
        from pyramid_twitterauth.view import _unpack_callback
        return _unpack_callback(*args, **kwargs)
    
    def assertRedirectedToFailed(self, response):
        from pyramid.httpexceptions import HTTPFound
        kwargs = dict(traverse=('failed',))
        self.request.route_url.assert_called_with('twitterauth', **kwargs)
        self.assertTrue(isinstance(response, HTTPFound))
    
    def test_denied(self):
        """Redirects to the failed view if the user denied our app."""
        
        self.request.GET['denied'] = 'abcd'
        response = self.makeOne(self.request)
        self.assertRedirectedToFailed(response)
    
    def test_convert_to_access_token_fails(self):
        """If our call to the Twitter API to convert our stored request token into
          an access token fails (for example, because Twitter is over capacity),
          redirects to the failed view.
        """
        
        import tweepy
        def raise_tweepy_error(*args, **kwargs):
            raise tweepy.TweepError('a')
        
        self.mock_handler.get_access_token = raise_tweepy_error
        response = self.makeOne(self.request, handler_factory=self.mock_handler_factory)
        self.assertRedirectedToFailed(response)
    
    def test_verify_credentials_fails(self):
        """If our call to the Twitter API to verify the user's credentials (and thus
          find out who they are) fails, redirects to the failed view.
        """
        
        import tweepy
        def raise_tweepy_error(*args, **kwargs):
            raise tweepy.TweepError('a')
        
        self.mock_client.verify_credentials = raise_tweepy_error
        response = self.makeOne(self.request, handler_factory=self.mock_handler_factory,
                Api=self.mock_api_factory)
        self.assertRedirectedToFailed(response)
    
    def test_verify_credentials_false(self):
        """Redirects to the failed view if the user's credentials aren't valid."""
        
        self.mock_client.verify_credentials.return_value = False
        response = self.makeOne(self.request, handler_factory=self.mock_handler_factory,
                Api=self.mock_api_factory)
        self.assertRedirectedToFailed(response)
    
    def test_returns_tuple_of_unpacked_elements(self):
        """Returns ``twitter_user, oauth_handler, access_permission``."""
        
        self.mock_client.verify_credentials.return_value = 'twitter user'
        self.mock_client.last_response.getheader.return_value = 'access permission'
        data = self.makeOne(self.request, handler_factory=self.mock_handler_factory,
                Api=self.mock_api_factory)
        self.assertTrue(data[0] == 'twitter user')
        self.assertTrue(data[1] == self.mock_handler)
        self.assertTrue(data[2] == 'access permission')
    

class TestAuthenticateCallback(unittest.TestCase):
    """Integration tests for ``view.authenticate_callback_view``.
    """
    
    def setUp(self):
        from mock import Mock
        from pyramid_twitterauth import view
        self._get_existing_twitter_account = view.get_existing_twitter_account
        self._get_redirect_url = view._get_redirect_url
        self._save_to_db = view.save_to_db
        self._remember = view.remember
        view.get_existing_twitter_account = Mock()
        view._get_redirect_url = Mock()
        view.save_to_db = Mock()
        view.remember = Mock()
        mock_request = Mock()
        self.request = mock_request
        self.request.session = {}
        self.request.is_authenticated = False
        self.mock_twitter_account = Mock()
        self.mock_twitter_user = Mock()
        self.mock_twitter_user.id = 1234
        self.mock_twitter_user.screen_name = 'screen name'
        self.mock_oauth_handler = Mock()
        self.mock_oauth_handler.access_token.key = 'key'
        self.mock_oauth_handler.access_token.secret = 'secret'
        self.access_permission = 'read'
        self.mock_unpack_callback = Mock()
        data = (self.mock_twitter_user, self.mock_oauth_handler, self.access_permission)
        self.mock_unpack_callback.return_value = data
        view._get_redirect_url.return_value = 'url'
        view.remember.return_value = {}
    
    def tearDown(self):
        from pyramid_twitterauth import view
        view.get_existing_twitter_account = self._get_existing_twitter_account
        view.save_to_db = self._save_to_db
        view.remember = self._remember
    
    def makeOne(self, request):
        """Call the OAuth callback view and return the response."""
        
        from pyramid_twitterauth.view import authenticate_callback_view
        return authenticate_callback_view(request, unpack=self.mock_unpack_callback)
    
    def test_authenticate_auth_user_forbidden(self):
        """Should not be called by an authenticated user."""
        
        from pyramid.httpexceptions import HTTPForbidden
        
        self.request.is_authenticated = True
        response = self.makeOne(self.request)
        self.assertTrue(isinstance(response, HTTPForbidden))
    
    def test_unpack_fails(self):
        """If unpack returns a redirect, returns it."""
        
        from pyramid.httpexceptions import HTTPFound
        
        redirect = HTTPFound(location='/foo')
        self.mock_unpack_callback.return_value = redirect
        response = self.makeOne(self.request)
        self.assertTrue(response.location == '/foo')
    
    def test_gets_existing_from_twitter_user_id(self):
        """Calls ``get_existing_twitter_account(twitter_user.id)``."""
        
        from pyramid_twitterauth import view
        
        self.mock_twitter_user.id = 1234
        response = self.makeOne(self.request)
        view.get_existing_twitter_account.assert_called_with(1234)
    
    def test_existing_fires_a_login_event(self):
        """If there's an existing twitter account, fires ``UserLoggedIn``."""
        
        from pyramid_twitterauth import view
        from pyramid_simpleauth.events import UserLoggedIn
        
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        event = self.request.registry.notify.call_args[0][0]
        self.assertTrue(isinstance(event, UserLoggedIn))
    
    def test_no_existing_fires_a_signup_event(self):
        """If there's not existing twitter account, fires ``UserSignedUp``."""
        
        from pyramid_twitterauth import view
        from pyramid_simpleauth.events import UserSignedUp
        
        view.get_existing_twitter_account.return_value = None
        response = self.makeOne(self.request)
        event = self.request.registry.notify.call_args[0][0]
        self.assertTrue(isinstance(event, UserSignedUp))
    
    def test_existing_saves_updated_twitter_account(self):
        """If there's an existing twitter account, its updated and saved."""
        
        from pyramid_twitterauth import view
        
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        view.save_to_db.assert_called_with(self.mock_twitter_account)
        
        self.assertTrue(self.mock_twitter_account.screen_name == 'screen name')
        self.assertTrue(self.mock_twitter_account.oauth_token == 'key')
        self.assertTrue(self.mock_twitter_account.oauth_token_secret == 'secret')
        self.assertTrue(self.mock_twitter_account.access_permission == 'read')
    
    def test_existing_doesnt_have_id_or_user_set(self):
        """If there's an existing twitter account, its id and user aren't set."""
        
        from pyramid_twitterauth import view
        from pyramid_simpleauth.model import User
        
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        self.assertFalse(isinstance(self.mock_twitter_account.user, User))
        self.assertFalse(self.mock_twitter_account.twitter_id == 1234)
    
    def test_no_existing_creates_and_saves(self):
        """If there's not an existing twitter account, inserts new 
          ``twitter_account`` and ``user``.
        """
        
        from pyramid_twitterauth import view
        from pyramid_simpleauth.model import User
        
        view.get_existing_twitter_account.return_value = None
        response = self.makeOne(self.request)
        twitter_account = view.save_to_db.call_args[0][0]
        
        self.assertTrue(twitter_account.twitter_id == 1234)
        self.assertTrue(twitter_account.screen_name == 'screen name')
        self.assertTrue(twitter_account.oauth_token == 'key')
        self.assertTrue(twitter_account.oauth_token_secret == 'secret')
        self.assertTrue(twitter_account.access_permission == 'read')
        self.assertTrue(isinstance(twitter_account.user, User))
        self.assertTrue(twitter_account.user.username == 'screen name')
    
    def test_logs_user_in(self):
        """Actually log the user in."""
        
        from pyramid_twitterauth import view
        
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        user = self.mock_twitter_account.user
        view.remember.assert_called_with(self.request, user.canonical_id)
    
    def test_existing_redirects_to_after_login(self):
        """If there's an existing twitter account, gets redirect url for login."""
        
        from pyramid_twitterauth import view
        
        self.request.session['twitter_oauth_next'] = 'next'
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        view._get_redirect_url.assert_called_with(self.request, 'login', 'next')
        self.assertTrue(response.location == 'url')
    
    def test_no_existing_redirects_to_after_signup(self):
        """If there's not an existing twitter account, gets redirect url for signup."""
        
        from pyramid_twitterauth import view
        
        self.request.session['twitter_oauth_next'] = 'next'
        view.get_existing_twitter_account.return_value = None
        response = self.makeOne(self.request)
        view._get_redirect_url.assert_called_with(self.request, 'signup', 'next')
        self.assertTrue(response.location == 'url')
    

class TestAuthorizeCallback(unittest.TestCase):
    """Integration tests for ``view.authorize_callback_view``.
    """
    
    def setUp(self):
        from mock import Mock
        from pyramid_simpleauth.model import User
        from pyramid_twitterauth import view
        self._get_existing_twitter_account = view.get_existing_twitter_account
        self._get_redirect_url = view._get_redirect_url
        self._save_to_db = view.save_to_db
        self._remember = view.remember
        view.get_existing_twitter_account = Mock()
        view._get_redirect_url = Mock()
        view.save_to_db = Mock()
        view.remember = Mock()
        mock_request = Mock()
        self.request = mock_request
        self.request.session = {}
        self.request.is_authenticated = True
        self.request.user = User()
        self.mock_twitter_account = Mock()
        self.mock_twitter_user = Mock()
        self.mock_twitter_user.id = 1234
        self.mock_twitter_user.screen_name = 'screen name'
        self.mock_oauth_handler = Mock()
        self.mock_oauth_handler.access_token.key = 'key'
        self.mock_oauth_handler.access_token.secret = 'secret'
        self.access_permission = 'read'
        self.mock_unpack_callback = Mock()
        data = (self.mock_twitter_user, self.mock_oauth_handler, self.access_permission)
        self.mock_unpack_callback.return_value = data
        view._get_redirect_url.return_value = 'url'
        view.remember.return_value = {}
    
    def tearDown(self):
        from pyramid_twitterauth import view
        view.get_existing_twitter_account = self._get_existing_twitter_account
        view.save_to_db = self._save_to_db
        view.remember = self._remember
    
    def makeOne(self, request):
        """Call the OAuth callback view and return the response."""
        
        from pyramid_twitterauth.view import authorize_callback_view
        return authorize_callback_view(request, unpack=self.mock_unpack_callback)
    
    def test_authenticate_auth_user_forbidden(self):
        """Should not be called after an authentication attempt."""
        
        from pyramid.httpexceptions import HTTPForbidden
        
        self.request.session['twitter_oauth_is_authenticate'] = True
        response = self.makeOne(self.request)
        self.assertTrue(isinstance(response, HTTPForbidden))
    
    def test_unpack_fails(self):
        """If unpack returns a redirect, returns it."""
        
        from pyramid.httpexceptions import HTTPFound
        
        redirect = HTTPFound(location='/foo')
        self.mock_unpack_callback.return_value = redirect
        response = self.makeOne(self.request)
        self.assertTrue(response.location == '/foo')
    
    def test_gets_existing_from_twitter_user_id(self):
        """Calls ``get_existing_twitter_account(twitter_user.id)``."""
        
        from pyramid_twitterauth import view
        
        self.mock_twitter_user.id = 1234
        response = self.makeOne(self.request)
        view.get_existing_twitter_account.assert_called_with(1234)
    
    def test_existing_saves_updated_twitter_account(self):
        """If there's an existing twitter account, its updated, related to
          ``request.user`` and saved.
        """
        
        from pyramid_twitterauth import view
        
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        view.save_to_db.assert_called_with(self.mock_twitter_account)
        
        self.assertTrue(self.mock_twitter_account.screen_name == 'screen name')
        self.assertTrue(self.mock_twitter_account.oauth_token == 'key')
        self.assertTrue(self.mock_twitter_account.oauth_token_secret == 'secret')
        self.assertTrue(self.mock_twitter_account.access_permission == 'read')
        self.assertTrue(self.mock_twitter_account.user == self.request.user)
    
    def test_no_existing_creates_and_saves(self):
        """If there's not an existing twitter account, creates one, sets
          the data, related to ``request.user`` and saves.
        """
        
        from pyramid_twitterauth import view
        
        view.get_existing_twitter_account.return_value = None
        response = self.makeOne(self.request)
        twitter_account = view.save_to_db.call_args[0][0]
        
        self.assertTrue(twitter_account.twitter_id == 1234)
        self.assertTrue(twitter_account.screen_name == 'screen name')
        self.assertTrue(twitter_account.oauth_token == 'key')
        self.assertTrue(twitter_account.oauth_token_secret == 'secret')
        self.assertTrue(twitter_account.access_permission == 'read')
        self.assertTrue(twitter_account.user == self.request.user)
    
    def test_redirects_to_after_connect(self):
        """Gets redirect url for connect."""
        
        from pyramid_twitterauth import view
        
        self.request.session['twitter_oauth_next'] = 'next'
        view.get_existing_twitter_account.return_value = self.mock_twitter_account
        response = self.makeOne(self.request)
        view._get_redirect_url.assert_called_with(self.request, 'connect', 'next')
        self.assertTrue(response.location == 'url')
    

class TestAuthenticateRedirects(unittest.TestCase):
    """Functional tests for the auth views."""
    
    def setUp(self):
        """Configure the Pyramid application."""
        
        self.config = config_factory()
        self.app = TestApp(self.config.make_wsgi_app())
    
    def tearDown(self):
        """Make sure the session is cleared between tests."""
        
        from pyramid_basemodel import Session
        Session.remove()
    
    def test_authenticate(self):
        """A request to the authenticate view should redirect to Twitter."""
        
        res = self.app.post('/oauth/twitter/authenticate', status=302)
        location = res.headers.get('Location')
        twitter_stub = 'https://api.twitter.com/oauth/authenticate?oauth_token='
        self.assertTrue(location.startswith(twitter_stub))
    
    def test_authorize(self):
        """A request to the authorize view will return 404."""
        
        res = self.app.post('/oauth/twitter/authorize', status=404)
    
    def test_invalid_consumer_settings(self):
        """A request to authenticate will redirect to failed if the consumer
          settings are invalid.
        """
        
        settings = self.config.registry.settings
        settings['twitterauth.oauth_consumer_key'] = 'blah'
        settings['twitterauth.oauth_consumer_secret'] = 'wrong'
        res = self.app.post('/oauth/twitter/authenticate', status=302)
        location = res.headers.get('Location')
        self.assertTrue(location=='http://localhost/oauth/twitter/failed')
    
    def test_authenticate_with_twitter_redirects_to_callback(self):
        """After a user authenticates with Twitter, they're redirected to the
          twitterauth callback view.
        """
        
        import urllib, urllib2
        
        # Get the auth url to redirect to.
        res = self.app.get('/oauth/twitter/authenticate', status=302)
        
        # Redirect and parse the response.
        sock = urllib2.urlopen(res.location)
        res = TestResponse()
        res.body = sock.read()
        sock.close()
        
        # Authenticate with our dummy username and password.
        res.form.set('session[username_or_email]', 'pyramid_twauth')
        res.form.set('session[password]', 'pyramid_twitterauth')
        
        # Submit the form and get the new redirect response.
        data = urllib.urlencode(res.form.submit_fields())
        sock = urllib2.urlopen(res.form.action, data)
        text = sock.read()
        sock.close()
        
        # Manually parse the redirect url out of the meta tag.
        l = len('<meta http-equiv="refresh" content="0;url=')
        pos = text.index('<meta http-equiv="refresh" content="0;url=')
        frag = text[pos+l:]
        pos = frag.index('">')
        url = frag[:pos]
        
        # Test that the fragment starts with our callback url.
        stub = 'http://localhost/oauth/twitter/authenticate_callback'
        self.assertTrue(url.startswith(stub))
    

class TestAuthorizeRedirects(unittest.TestCase):
    """Functional tests for the authorize workflow."""
    
    def setUp(self, is_authenticated=True):
        """Configure the Pyramid application."""
        
        settings = {'twitterauth.mode': 'connect'}
        self.config = config_factory(is_authenticated=is_authenticated, **settings)
        self.app = TestApp(self.config.make_wsgi_app())
    
    def tearDown(self):
        """Make sure the session is cleared between tests."""
        
        from pyramid_basemodel import Session
        Session.remove()
    
    def test_authenticate(self):
        """A request to the authenticate view will return 404."""
        
        res = self.app.post('/oauth/twitter/authenticate', status=404)
    
    def test_authorize(self):
        """A request to the authorize view requires authentication and will
          redirect to Twitter.
        """
        
        self.setUp(is_authenticated=False)
        res = self.app.post('/oauth/twitter/authorize', status=302)
        location = res.headers.get('Location')
        auth_stub = 'http://localhost/auth/login?next'
        self.assertTrue(location.startswith(auth_stub))
        
        self.setUp(is_authenticated=True)
        res = self.app.post('/oauth/twitter/authorize', status=302)
        location = res.headers.get('Location')
        twitter_stub = 'https://api.twitter.com/oauth/authorize?oauth_token='
        self.assertTrue(location.startswith(twitter_stub))
    
    def test_invalid_consumer_settings(self):
        """A request to authorize will redirect to failed if the consumer
          settings are invalid.
        """
        
        settings = self.config.registry.settings
        settings['twitterauth.oauth_consumer_key'] = 'blah'
        settings['twitterauth.oauth_consumer_secret'] = 'wrong'
        res = self.app.post('/oauth/twitter/authorize', status=302)
        location = res.headers.get('Location')
        self.assertTrue(location=='http://localhost/oauth/twitter/failed')
    
    def test_authorize_with_twitter_redirects_to_callback(self):
        """After a user authorizes with Twitter, they're redirected to the
          authorize_callback view.
        """
        
        import urllib, urllib2
        
        # Get the auth url to redirect to.
        res = self.app.get('/oauth/twitter/authorize', status=302)
        
        # Redirect and parse the response.
        sock = urllib2.urlopen(res.location)
        res = TestResponse()
        res.body = sock.read()
        sock.close()
        
        # Authenticate with our dummy username and password.
        res.form.set('session[username_or_email]', 'pyramid_twauth')
        res.form.set('session[password]', 'pyramid_twitterauth')
        
        # Submit the form and get the new redirect response.
        data = urllib.urlencode(res.form.submit_fields())
        sock = urllib2.urlopen(res.form.action, data)
        text = sock.read()
        sock.close()
        
        # Manually parse the redirect url out of the meta tag.
        l = len('<meta http-equiv="refresh" content="0;url=')
        pos = text.index('<meta http-equiv="refresh" content="0;url=')
        frag = text[pos+l:]
        pos = frag.index('">')
        url = frag[:pos]
        
        # Test that the fragment starts with our callback url.
        stub = 'http://localhost/oauth/twitter/authorize_callback'
        self.assertTrue(url.startswith(stub))
    




# -*- coding: utf-8 -*-

"""Provides authenticate, authorize and callback views and a failed view to
  redirect to when OAuth fails, e.g.: when Twitter is down.
"""

import logging
logger = logging.getLogger(__name__)

import tweepy

from pyramid.httpexceptions import HTTPForbidden, HTTPFound, HTTPUnauthorized
from pyramid.security import forget, remember, unauthenticated_userid
from pyramid.security import NO_PERMISSION_REQUIRED as PUBLIC
from zope.interface.registry import ComponentLookupError

from pyramid_basemodel import save as save_to_db
from pyramid_simpleauth.events import UserSignedUp, UserLoggedIn
from pyramid_simpleauth.schema import Invalid, RequestPath
from pyramid_simpleauth.model import User

from .hooks import get_handler
from .model import get_existing_twitter_account, TwitterAccount, TwitterProfile

def forbidden_view(request):
    """Handle a user being denied access to a resource or view by redirecting
      to authenticate via Twitter.  See the ``pyramid_twitterauth.includeme``
      function for info on how to expose this view.
      
      Setup::
      
          >>> from mock import Mock
          >>> from pyramid_twitterauth import view
          >>> _unauthenticated_userid = view.unauthenticated_userid
          >>> view.unauthenticated_userid = Mock()
          >>> mock_request = Mock()
          >>> mock_request.path = '/forbidden/page'
          >>> mock_request.route_url.return_value = '/oauth/twitter/authenticate'
      
      If the user is already logged in, it means they don't have the requisit
      permission, so we raise a 403 Forbidden error::
      
          >>> view.unauthenticated_userid.return_value = 1234
          >>> response = forbidden_view(mock_request)
          >>> response.status
          '403 Forbidden'
      
      Otherwise we redirect to the authenticate view::
      
          >>> view.unauthenticated_userid.return_value = None
          >>> response = forbidden_view(mock_request)
          >>> kwargs = {
          ...     '_query': (('next', '/forbidden/page'),),
          ...     'traverse': ('authenticate',)
          ... }
          >>> mock_request.route_url.assert_called_with('twitterauth', **kwargs)
          >>> response.location
          '/oauth/twitter/authenticate'
          >>> response.status
          '302 Found'
      
      Teardown::
      
          >>> view.unauthenticated_userid = _unauthenticated_userid
      
    """
    
    if unauthenticated_userid(request):
        return HTTPForbidden()
    query = (('next', request.path),)
    url = request.route_url('twitterauth', traverse=('authenticate',), _query=query)
    return HTTPFound(location=url)


def _redirect_to_failed(request, redirect_cls=HTTPFound):
    """Redirect to the failed view.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_request = Mock()
          >>> mock_request.route_url.return_value = 'redirect url'
          >>> mock_redirect_cls = Mock()
          >>> mock_redirect_cls.return_value = 'http found'
      
      Test::
      
          >>> _redirect_to_failed(mock_request, redirect_cls=mock_redirect_cls)
          'http found'
          >>> kwargs = dict(traverse=('failed',))
          >>> mock_request.route_url.assert_called_with('twitterauth', **kwargs)
          >>> mock_redirect_cls.assert_called_with(location='redirect url')
      
    """
    
    url = request.route_url('twitterauth', traverse=('failed',))
    return redirect_cls(location=url)

def _do_oauth_redirect(request, is_authenticate, handler_factory=get_handler):
    """Start the OAuth dance by getting a request token from the Twitter API,
      storing it in the session and then redirecting to Twitter.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_request = Mock()
          >>> mock_request.session = {}
          >>> mock_handler = Mock()
          >>> mock_handler.get_authorization_url.return_value = 'url'
          >>> mock_handler.request_token.key = 'key'
          >>> mock_handler.request_token.secret = 'secret'
          >>> mock_handler_factory = Mock()
          >>> mock_handler_factory.return_value = mock_handler
      
      Stores whether this OAuth attempt is a signin or not::
      
          >>> return_value = _do_oauth_redirect(mock_request, True,
          ...         handler_factory=mock_handler_factory)
          >>> mock_request.session['twitter_oauth_is_authenticate']
          True
      
      If there's a ``next`` param in the request, stores that too::
      
          >>> mock_request.params = {'next': '/foo/bar'}
          >>> return_value = _do_oauth_redirect(mock_request, True,
          ...         handler_factory=mock_handler_factory)
          >>> mock_request.session.get('twitter_oauth_next')
          u'/foo/bar'
      
      As long as it's a valid path::
      
          >>> mock_request.session = {}
          >>> mock_request.params = {'next': '<script src="javascript:h@x();">'}
          >>> return_value = _do_oauth_redirect(mock_request, True,
          ...         handler_factory=mock_handler_factory)
          >>> mock_request.session.get('twitter_oauth_next')
      
      Gets a request token from Twitter::
      
          >>> mock_request.params = {}
          >>> mock_request.session = {}
          >>> return_value = _do_oauth_redirect(mock_request, True,
          ...         handler_factory=mock_handler_factory)
          >>> kwargs = dict(callback='authenticate_callback')
          >>> mock_handler_factory.assert_called_with(mock_request, **kwargs)
          >>> kwargs = dict(signin_with_twitter=True)
          >>> mock_handler.get_authorization_url.assert_called_with(**kwargs)
      
      Stores it in the session::
      
          >>> mock_request.session.get('twitter_request_token_key')
          'key'
          >>> mock_request.session.get('twitter_request_token_secret')
          'secret'
      
      Redirects to the authorisation url::
      
          >>> return_value.location
          'url'
      
    """
    
    # Store whether this OAuth attempt is to authenticate or authorize, so we
    # can provide the right try again link in case any of our calls to the
    # Twitter API fail.
    request.session['twitter_oauth_is_authenticate'] = is_authenticate
    # If there was a valid ``next`` param in the query string, store it in the
    # session so we can redirect to it later on.
    next_ = request.params.get('next')
    try:
        request.session['twitter_oauth_next'] = RequestPath.to_python(next_)
    except Invalid:
        pass
    # Initialise an OAuth handler with the right consumer settings.
    if is_authenticate:
        callback_view = 'authenticate_callback'
    else:
        callback_view = 'authorize_callback'
    oauth_handler = handler_factory(request, callback=callback_view)
    # Try and get a request token from Twitter.
    kwargs = dict(signin_with_twitter=is_authenticate)
    try:
        redirect_url = oauth_handler.get_authorization_url(**kwargs)
    except tweepy.TweepError:
        return _redirect_to_failed(request)
    # Store the request token in the session.
    token = oauth_handler.request_token
    request.session['twitter_request_token_key'] = token.key
    request.session['twitter_request_token_secret'] = token.secret
    # Redirect.
    return HTTPFound(location=redirect_url)

def _get_redirect_url(request, action, candidate_next, user=None):
    """Get URL to redirect to after ``action``.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_request = Mock()
          >>> mock_request.route_url.return_value = 'url'
          >>> mock_request.registry.settings.get.return_value = 'route name'
      
      Called with a valid next, returns it::
      
          >>> _get_redirect_url(mock_request, 'login', '/bar') # Valid
          u'/bar'
      
      Called with an invalid next doesn't::
      
          >>> _get_redirect_url(mock_request, 'login', '://') == u'://'
          False
      
      If action is 'login', tries to get the ``after_login_route``::
      
          >>> return_value = _get_redirect_url(mock_request, 'login', None)
          >>> get = mock_request.registry.settings.get
          >>> get.assert_called_with('simpleauth.after_login_route', 'index')
          >>> mock_request.route_url.assert_called_with('route name', traverse=())
          >>> return_value
          'url'
      
      With the user.username in the traverse if a user is provided::
      
          >>> mock_user = Mock()
          >>> _ = _get_redirect_url(mock_request, 'login', None, user=mock_user)
          >>> mock_request.route_url.assert_called_with('route name', 
          ...         traverse=(mock_user.username,))
      
      If action is 'signup', tries to get the ``after_signup_route``::
      
          >>> return_value = _get_redirect_url(mock_request, 'signup', None)
          >>> get = mock_request.registry.settings.get
          >>> get.assert_called_with('simpleauth.after_signup_route', 'users')
      
      If 'connect', tries to get the ``after_connect_route``::
      
          >>> return_value = _get_redirect_url(mock_request, 'connect', None)
          >>> get = mock_request.registry.settings.get
          >>> get.assert_called_with('twitterauth.after_connect_route', 'users')
      
      If the route isn't configured, falls back on '/'::
      
          >>> def raise_err(*args, **kwargs):
          ...     raise ComponentLookupError
          ... 
          >>> mock_request.route_url = raise_err
          >>> _get_redirect_url(mock_request, 'signup', None)
          '/'
      
    """
    
    # If we've been provided a valid next value, return that.
    if candidate_next:
        try:
            return RequestPath.to_python(candidate_next)
        except Invalid:
            pass
    # Otherwise get the route url to redirect to.
    settings = request.registry.settings
    if action == 'login':
        key = 'simpleauth.after_login_route'
        default = 'index'
    elif action == 'signup':
        key = 'simpleauth.after_signup_route'
        default = 'users'
    elif action == 'connect':
        key = 'twitterauth.after_connect_route'
        default = 'users'
    route_name = settings.get(key, default)
    traverse = (user.username,) if user else ()
    try:
        # If the route exists, use that URL.
        return request.route_url(route_name, traverse=traverse)
    except (KeyError, ComponentLookupError):
        # Otherwise as a last resort fall back on '/'.
        return '/'

def _unpack_callback(request, handler_factory=get_handler, Api=tweepy.API):
    """Get a Twitter user, OAuth handler and access permission from an OAuth
      callback request.  Returns a tuple of::
      
          twitter_user, oauth_handler, access_permission
      
      Handles the following failure scenarios:
      
      * the user denied our app
      * our call to the Twitter API to convert our stored request token into
        an access token fails (for example, because Twitter is over capacity)
      * our call to the Twitter API to verify the user's credentials (and thus
        find out who they are) fails
      * the user's credentials are not verified, i.e.: for some strange reason
        the access token we got doesn't actually work
      
    """
    
    # If the user chose not to authorize the app, redirect to the failed view.
    if request.GET.get('denied'):
        return _redirect_to_failed(request)
    # Get the verifier and / or denied values from the request query string
    # and pop the request token from the session, where we saved it earlier.
    verifier = request.GET.get('oauth_verifier')
    request_token_key = request.session['twitter_request_token_key']
    request_token_secret = request.session['twitter_request_token_secret']
    del request.session['twitter_request_token_key']
    del request.session['twitter_request_token_secret']
    # Exchange the request token for an access token.
    oauth_handler = handler_factory(request)
    oauth_handler.set_request_token(request_token_key, request_token_secret)
    try:
        oauth_handler.get_access_token(verifier)
    except tweepy.TweepError:
        return _redirect_to_failed(request)
    # Get the authenticated user's Twitter details.
    client = Api(oauth_handler)
    try:
        twitter_user = client.verify_credentials()
    except tweepy.TweepError:
        return _redirect_to_failed(request)
    # If the credentials were not verified, redirect to the failed view.
    if twitter_user is False:
        return _redirect_to_failed(request)
    # Get the access permission level from the last_response.
    access_permission = client.last_response.getheader('X-Access-Level')
    # Return unpacked elements.
    return twitter_user, oauth_handler, access_permission


def oauth_authenticate_view(request, do_redirect=_do_oauth_redirect):
    """Redirect to GET oauth/authenticate.
      
      Setup::
      
          >>> from mock import Mock
          >>> from pyramid_twitterauth import view
          >>> _get_redirect_url = view._get_redirect_url
          >>> view._get_redirect_url = Mock()
          >>> view._get_redirect_url.return_value = 'redirect url'
          >>> mock_request = Mock()
          >>> mock_request.is_authenticated = False
          >>> mock_do_redirect = Mock()
          >>> mock_do_redirect.return_value = 'twitter api'
      
      Redirects to Twitter::
      
          >>> oauth_authenticate_view(mock_request, do_redirect=mock_do_redirect)
          'twitter api'
          >>> mock_do_redirect.assert_called_with(mock_request, True)
      
      Unless called with an authenticated user::
      
          >>> mock_do_redirect = Mock()
          >>> mock_request.is_authenticated = True
          >>> oauth_authenticate_view(mock_request, do_redirect=mock_do_redirect)
          'redirect url'
          >>> mock_do_redirect.called
          False
      
      Teardown::
      
          >>> view._get_redirect_url = _get_redirect_url
      
    """
    
    # If there's already an authenticated user, we don't need to authenticate.
    if request.is_authenticated:
        return _get_redirect_url(request, 'login', request.params.get('next'))
    return do_redirect(request, True)

def oauth_authorize_view(request, do_redirect=_do_oauth_redirect):
    """Redirect to GET oauth/authorize.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_request = Mock()
          >>> mock_do_redirect = Mock()
          >>> mock_do_redirect.return_value = 'http found'
      
      Test::
      
          >>> oauth_authorize_view(mock_request, do_redirect=mock_do_redirect)
          'http found'
          >>> mock_do_redirect.assert_called_with(mock_request, False)
      
    """
    
    return do_redirect(request, False)


def authenticate_callback_view(request, unpack=_unpack_callback):
    """Complete the OAuth dance after a user has authenticated the app."""
    
    # This view should not be called by an authenticated user.
    if request.is_authenticated:
        return HTTPForbidden()
    
    # Unpack the request.
    return_value = unpack(request)
    if isinstance(return_value, HTTPFound):
        return return_value
    twitter_user, oauth_handler, access_permission = return_value
    # If there is an existing ``twitter_account`` then this is a login, so
    # update the ``twitter_account`` and generate a login event.
    existing = get_existing_twitter_account(twitter_user.id)
    if existing:
        twitter_account = existing
        twitter_account.profile.set_data_from_tweepy_user(twitter_user)
        user = twitter_account.user
        event = UserLoggedIn(request, user, data=twitter_user)
        action = 'login'
    else: # Otherwise, this is a signup, so insert a new ``user`` with a
        # ``twitter_account`` and generate a signup event.
        user = User()
        user.username = twitter_user.screen_name
        twitter_account = TwitterAccount()
        twitter_account.twitter_id = twitter_user.id
        twitter_account.user = user
        twitter_account.profile = TwitterProfile.create_from_tweepy_user(twitter_user)
        event = UserSignedUp(request, user, data=twitter_user)
        action = 'signup'
    # Update the twitter_account with the latest data, save to the db and
    # actually fire the event.
    twitter_account.screen_name = twitter_user.screen_name
    twitter_account.oauth_token = oauth_handler.access_token.key
    twitter_account.oauth_token_secret = oauth_handler.access_token.secret
    twitter_account.access_permission = access_permission
    save_to_db(twitter_account) # <!-- this saves the user along with it.
    request.registry.notify(event)
    # Actually log the user in and then redirect to the appropriate location.
    next_ = request.session.get('twitter_oauth_next')
    location = _get_redirect_url(request, action, next_, user=user)
    headers = remember(request, user.canonical_id)
    return HTTPFound(location=location, headers=headers)

def authorize_callback_view(request, unpack=_unpack_callback):
    """Complete the OAuth dance after a user has authorized the app."""
    
    # This callback should not be called to handle login or signup.
    if request.session.get('twitter_oauth_is_authenticate'):
        return HTTPForbidden()
    # Unpack the request.
    return_value = unpack(request)
    if isinstance(return_value, HTTPFound):
        return return_value
    twitter_user, oauth_handler, access_permission = return_value
    # Update or create the ``twitter_account`` corresponding to ``twitter_user``,
    # relate it to the current ``request.user`` and save to the db.
    twitter_account = get_existing_twitter_account(twitter_user.id)
    if not twitter_account:
        twitter_account = TwitterAccount()
        twitter_account.twitter_id = twitter_user.id
    twitter_account.screen_name = twitter_user.screen_name
    twitter_account.oauth_token = oauth_handler.access_token.key
    twitter_account.oauth_token_secret = oauth_handler.access_token.secret
    twitter_account.access_permission = access_permission
    twitter_account.user = request.user
    save_to_db(twitter_account)
    # Redirect to the appropriate location.
    next_ = request.session.get('twitter_oauth_next')
    location = _get_redirect_url(request, 'connect', next_, user=request.user)
    return HTTPFound(location=location)


def oauth_failed_view(request):
    """Render a page explaining that Twitter Auth failed, with a link to try
      again, e.g.: after a few seconds when Twitter is back up and working.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_request = Mock()
      
      Redirects to authenticate if is a signin and adds a ``next`` param to the
      redirect url if stored in the session::
      
          >>> mock_request.session = {
          ...     'twitter_oauth_is_authenticate': True,
          ...     'twitter_oauth_next': '/next'
          ... }
          >>> return_value = oauth_failed_view(mock_request)
          >>> kwargs = {
          ...     'traverse': ('authenticate',),
          ...     'query': (('next', '/next'),)
          ... }
          >>> mock_request.route_url.assert_called_with('twitterauth', **kwargs)
      
      Redirects to authorize if not a signin::
      
          >>> mock_request.session = {
          ...     'twitter_oauth_is_authenticate': False,
          ...     'twitter_oauth_next': None
          ... }
          >>> return_value = oauth_failed_view(mock_request)
          >>> kwargs = {
          ...     'traverse': ('authorize',),
          ...     'query': ()
          ... }
          >>> mock_request.route_url.assert_called_with('twitterauth', **kwargs)
      
    """
    
    # Work out which view to link to.
    is_authenticate = request.session.get('twitter_oauth_is_authenticate')
    traverse = ('authenticate',) if is_authenticate else ('authorize',)
    
    # If stored in the session, add a ``next`` param.
    next_url = request.session.get('twitter_oauth_next')
    query = (('next', next_url),) if next_url else ()
    
    # Generate the try again url.
    url = request.route_url('twitterauth', traverse=traverse, query=query)
    return {'try_again_url': url}




# -*- coding: utf-8 -*-

"""Provides an SQLAlchemy based ``TwitterAccount`` model class."""

import json

from sqlalchemy import Column, ForeignKey
from sqlalchemy import BigInteger, Integer, Unicode, UnicodeText
from sqlalchemy.orm import backref, relationship

from zope.interface import implements

from pyramid_basemodel import Base, BaseMixin, Session, save
from pyramid_simpleauth import model as simpleauth_model

_user_backref = backref("twitter_account", lazy="joined", uselist=False)
_account_backref = backref("profile", uselist=False)

from .interfaces import ITwitterAccount

class TwitterAccount(Base, BaseMixin):
    """A user's twitter account with oauth token and access permission data."""
    
    implements(ITwitterAccount)
    
    __tablename__ = 'auth_twitter_accounts'
    
    twitter_id = Column(BigInteger, unique=True)
    screen_name = Column(Unicode(20))
    
    oauth_token = Column(Unicode(200))
    oauth_token_secret = Column(Unicode(200))
    access_permission = Column(Unicode(64))
    
    # XXX defines a relation to a table defined in `pyramid_simpleauth`.
    user_id = Column(Integer, ForeignKey('auth_users.id'))
    user = relationship(simpleauth_model.User, lazy='joined', backref=_user_backref)
    
    def __json__(self):
        """Return a dictionary representation of the ``TwitterAccount`` instance.
          
              >>> account = TwitterAccount(twitter_id=1234, screen_name='thruflo')
              >>> account.__json__()
              {'twitter_id': 1234, 'screen_name': 'thruflo'}
          
        """
        
        data = {'twitter_id': self.twitter_id, 'screen_name': self.screen_name}
        if self.profile:
            data.update(self.profile.data)
        return data
    

class TwitterProfile(Base, BaseMixin):
    """Stores a user's twitter profile data as a string."""
    
    __tablename__ = 'auth_twitter_profiles'
    
    id = Column(BigInteger, ForeignKey('auth_twitter_accounts.twitter_id'),
            primary_key=True)
    twitter_account = relationship(TwitterAccount, backref=_account_backref)
    
    @property
    def data(self):
        return json.loads(self.data_str)
    
    data_str = Column(UnicodeText)
    
    def set_data_from_tweepy_user(self, user):
        data = user.__getstate__()
        if 'status' in data:
            del data['status']
        if 'created_at' in data:
            data['created_at'] = data['created_at'].isoformat()
        self.data_str = json.dumps(data)
    
    @classmethod
    def create_from_tweepy_user(cls, user):
        profile = TwitterProfile()
        profile.id = user.id
        profile.set_data_from_tweepy_user(user)
        return profile
    


def get_existing_twitter_account(twitter_id, cls=TwitterAccount):
    """Get an existing twitter account from the ``twitter_id`` provided.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_cls = Mock()
          >>> mock_filtered_query = Mock()
          >>> mock_filtered_query.first.return_value = 'twitter user 1'
          >>> mock_cls.query.filter_by.return_value = mock_filtered_query
      
      Queries using the ``twitter_id`` and returns the first result::
      
          >>> get_existing_twitter_account(1, cls=mock_cls)
          'twitter user 1'
          >>> mock_cls.query.filter_by.assert_called_with(twitter_id=1)
      
    """
    
    query = cls.query.filter_by(twitter_id=twitter_id)
    return query.first()




# -*- coding: utf-8 -*-

"""Provides ``TwitterRequestAPI`` class and a ``get_twitter()`` function to get
  an instance of it from the current ``request``.
"""

import tweepy

def get_handler(request, callback=None, Handler=tweepy.OAuthHandler):
    """Convienience function to get an appropriately configured ``Handler``
      instance from the current request.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_request = Mock()
          >>> mock_request.route_url.return_value = 'callback'
          >>> mock_request.registry.settings = {
          ...     'twitterauth.oauth_consumer_key': 'key',
          ...     'twitterauth.oauth_consumer_secret': 'secret'
          ... }
          >>> mock_handler_cls = Mock()
          >>> mock_handler_cls.return_value = 'configured handler'
      
      Returns a ``Handler`` initialised with the consumer settings and the
      callback url::
      
          >>> get_handler(mock_request, Handler=mock_handler_cls)
          'configured handler'
          >>> args = ('key', 'secret')
          >>> kwargs = {'callback': None, 'secure': True}
          >>> mock_handler_cls.assert_called_with(*args, **kwargs)
      
      Constructing the callback url from the ``twitterauth`` route::
      
          >>> get_handler(mock_request, callback='callback', Handler=mock_handler_cls)
          'configured handler'
          >>> mock_request.route_url.assert_called_with('twitterauth', 
          ...         traverse=('callback',))
      
    """
    
    # Get the Twitter app consumer settings.
    settings = request.registry.settings
    key = settings.get('twitterauth.oauth_consumer_key')
    secret = settings.get('twitterauth.oauth_consumer_secret')
    # Construct the callback url for the oauth dance.
    if callback:
        callback = request.route_url('twitterauth', traverse=(callback,))
    # Return the configured handler.
    return Handler(key, secret, callback=callback, secure=True)


class TwitterRequestAPI(object):
    """Adapts a ``request`` to provide an authenticated Twitter api client
      and information on the Twitter api access level we have for that request's
      authenticated user.
    """
    
    client = None
    access_permission = ''
    
    @property
    def has_read_access(self):
        """Do we have read access to the Twitter api for ``self.user``?
          
          Setup::
          
              >>> from mock import Mock
              >>> mock_request = Mock()
              >>> mock_request.user = None
          
          If ``self.access_permission`` contains 'read', returns ``True``::
          
              >>> twitter = TwitterRequestAPI(mock_request)
              >>> twitter.access_permission = 'read'
              >>> twitter.has_read_access
              True
          
          Else returns ``False``::
          
              >>> twitter.access_permission = 'Flobble'
              >>> twitter.has_write_access
              False
          
        """
        
        return 'read' in self.access_permission
    
    @property
    def has_write_access(self):
        """Do we have write access to the Twitter api for ``self.user``?
          
          Setup::
          
              >>> from mock import Mock
              >>> mock_request = Mock()
              >>> mock_request.user = None
          
          If ``self.access_permission`` contains 'write', returns ``True``::
          
              >>> twitter = TwitterRequestAPI(mock_request)
              >>> twitter.access_permission = 'read-write'
              >>> twitter.has_write_access
              True
          
          Else returns ``False``::
          
              >>> twitter.access_permission = 'Flobble'
              >>> twitter.has_write_access
              False
          
        """
        
        return 'write' in self.access_permission
    
    
    def __init__(self, request, handler_factory=get_handler, Api=tweepy.API):
        """Initialise an OAuth handler with the right consumer settings, then
          if the user has a twitter account setup the twitter api client and
          set the access permission.
          
          Setup::
          
              >>> from mock import Mock
              >>> mock_request = Mock()
              >>> mock_user = Mock()
          
          If there isn't an authenticated user, no dice::
          
              >>> mock_request.user = None
              >>> twitter = TwitterRequestAPI(mock_request)
              >>> twitter.client
              >>> twitter.has_read_access
              False
          
          Ditto a user without a Twitter account::
          
              >>> mock_user.twitter_account = None
              >>> mock_request.user = mock_user
              >>> twitter = TwitterRequestAPI(mock_request)
              >>> twitter.client
              >>> twitter.has_read_access
              False
          
          More setup::
          
              >>> mock_twitter_account = Mock()
              >>> mock_twitter_account.oauth_token = 'token'
              >>> mock_twitter_account.oauth_token_secret = 'token secret'
              >>> mock_twitter_account.access_permission = 'read'
              >>> mock_handler = Mock()
              >>> mock_handler_factory = Mock()
              >>> mock_handler_factory.return_value = mock_handler
              >>> mock_api_factory = Mock()
              >>> mock_api_factory.return_value = '<api client>'
          
          If we do have an authenticated user with a Twitter account, then we
          authenticate the oauth handler::
          
              >>> mock_user.twitter_account = mock_twitter_account
              >>> twitter = TwitterRequestAPI(mock_request, 
              ...                             handler_factory=mock_handler_factory, 
              ...                             Api=mock_api_factory)
              >>> mock_handler_factory.assert_called_with(mock_request)
              >>> mock_handler.set_access_token.assert_called_with('token', 
              ...                                                  'token secret')
              >>> mock_api_factory.assert_called_with(mock_handler)
              >>> twitter.client
              '<api client>'
              >>> twitter.has_read_access
              True
          
        """
        
        # Initialise an OAuth handler with the right consumer settings.
        oauth_handler = handler_factory(request)
        # If the user has a twitter account.
        if request.user and hasattr(request.user, 'twitter_account'):
            twitter_account = request.user.twitter_account
            if twitter_account:
                # Authenticate the handler with the user's access token details.
                oauth_token = twitter_account.oauth_token
                oauth_token_secret = twitter_account.oauth_token_secret
                oauth_handler.set_access_token(oauth_token, oauth_token_secret)
                # Setup the twitter api client
                self.client = Api(oauth_handler)
                # Set the access permission
                self.access_permission = twitter_account.access_permission
    


# Reified as the ``request.twitter`` property.
def get_twitter(request, cls=TwitterRequestAPI):
    """Return a ``cls`` instance instantiated with ``request``::
      
          >>> from mock import Mock
          >>> mock_cls = Mock()
          >>> mock_cls.return_value = '<twitter instance>'
          >>> mock_request = Mock()
          >>> get_twitter(mock_request, cls=mock_cls)
          '<twitter instance>'
          >>> mock_cls.assert_called_with(mock_request)
      
    """
    
    return cls(request)




# -*- coding: utf-8 -*-

from pyramid.httpexceptions import HTTPForbidden
from pyramid.security import NO_PERMISSION_REQUIRED as PUBLIC

from .hooks import get_twitter
from .view import forbidden_view, oauth_failed_view
from .view import oauth_authenticate_view, oauth_authorize_view
from .view import authenticate_callback_view, authorize_callback_view

def includeme(config):
    """Allow developers to use ``config.include('pyramid_twitterauth')``.
      
      Setup::
      
          >>> from mock import Mock
          >>> mock_config = Mock()
          >>> mock_config.registry.settings = {}
      
      Adds ``twitter`` property to the ``request``::
      
          >>> includeme(mock_config)
          >>> args = (get_twitter, 'twitter')
          >>> mock_config.set_request_property.assert_any_call(*args, reify=True)
      
      Exposes the ``twitterauth`` route::
      
          >>> args = ('twitterauth', 'oauth/twitter/*traverse')
          >>> mock_config.add_route.assert_any_call(*args)
      
      Exposes the ``oauth_failed_view``::
      
          >>> kwargs = dict(route_name="twitterauth", name='failed', permission=PUBLIC,
          ...         renderer='pyramid_twitterauth:templates/failed.mako')
          >>> mock_config.add_view.assert_any_call(oauth_failed_view, **kwargs)
      
      If in connect mode, exposes the ``authorize`` views::
      
          >>> mock_config = Mock()
          >>> mock_config.registry.settings = {}
          >>> mock_config.registry.settings['twitterauth.mode'] = 'connect'
          >>> includeme(mock_config)
          >>> args = (oauth_authorize_view,)
          >>> kwargs = dict(route_name="twitterauth", name='authorize')
          >>> mock_config.add_view.assert_any_call(*args, **kwargs)
          >>> args = (authorize_callback_view,)
          >>> kwargs = dict(route_name="twitterauth", name='authorize_callback')
          >>> mock_config.add_view.assert_any_call(*args, **kwargs)
      
      But not the ``authenticate`` views::
      
          >>> args = (oauth_authenticate_view,)
          >>> kwargs = dict(route_name="twitterauth", name='authenticate', 
          ...         permission=PUBLIC)
          >>> mock_config.add_view.assert_any_call(*args, **kwargs) # doctest: +ELLIPSIS
          Traceback (most recent call last):
          ...
          AssertionError: ...
      
      Otherwise, exposes the ``authenticate`` views and adds the forbidden view.
      (Note that this may require a call ``config.commit()`` inbetween including
      ``pyramid_simpelauth`` and ``pyramid_twitterauth`` to manually resolve a
      ``pyramid.exceptions.ConfigurationConflictError``)::
      
          >>> mock_config = Mock()
          >>> mock_config.registry.settings = {}
          >>> includeme(mock_config)
          >>> kwargs = dict(context=HTTPForbidden, permission=PUBLIC)
          >>> mock_config.add_view.assert_any_call(forbidden_view, **kwargs)
          >>> args = (oauth_authenticate_view,)
          >>> kwargs = dict(route_name="twitterauth", name='authenticate', 
          ...         permission=PUBLIC)
          >>> mock_config.add_view.assert_any_call(*args, **kwargs)
          >>> args = (authenticate_callback_view,)
          >>> kwargs = dict(route_name="twitterauth",
          ...         name='authenticate_callback', permission=PUBLIC)
          >>> mock_config.add_view.assert_any_call(*args, **kwargs)
      
      And not the ``authorize`` views::
      
          >>> args = (oauth_authorize_view,)
          >>> kwargs = dict(route_name="twitterauth", name='authorize')
          >>> mock_config.add_view.assert_any_call(*args, **kwargs) # doctest: +ELLIPSIS
          Traceback (most recent call last):
          ...
          AssertionError: ...
      
    """
    
    # Add ``is_authenticated`` and ``user`` properties to the request.
    settings = config.registry.settings
    config.set_request_property(get_twitter, 'twitter', reify=True)
    
    # Expose the ``twitterauth`` route.
    prefix = settings.get('twitterauth.url_prefix', 'oauth/twitter')
    config.add_route('twitterauth', '{0}/*traverse'.format(prefix))
    
    # Expose the ``failed`` view.
    config.add_view(oauth_failed_view, route_name="twitterauth", name='failed',
            permission=PUBLIC, renderer='pyramid_twitterauth:templates/failed.mako')
    
    # If we're in connect mode, which means we want to allow existing users to
    # connect their Twitter accounts, expose the ``oauth_authorize`` and 
    # ``authorize_callback`` views.
    if settings.get('twitterauth.mode') == 'connect':
        config.add_view(oauth_authorize_view, route_name="twitterauth", name='authorize')
        config.add_view(authorize_callback_view, route_name="twitterauth",
                name='authorize_callback')
    # Otherwise, expose ``oauth_authenticate`` and ``authenticate_callback`` which
    # allow users to signup and login exclusively through Twitter and expose the
    # ``forbidden_view`` so that unauthenticated users accessing a protected
    # resource are automatically redirected to ``oauth_authenticate``.
    else:
        config.add_view(forbidden_view, context=HTTPForbidden, permission=PUBLIC)
        config.add_view(oauth_authenticate_view, route_name="twitterauth",
                name='authenticate', permission=PUBLIC)
        config.add_view(authenticate_callback_view, route_name="twitterauth",
                name='authenticate_callback', permission=PUBLIC)
    

